@@ -86,6 +86,9 @@ move the .new to the plain snapshot name as a means of recovery.
 	db_file_base/2,			% DB, FileBase
 	file_base_db/2.			% FileBase, DB
 
+:- meta_predicate
+	no_agc(0).
+
 %%	rdf_attach_db(+Directory, +Options)
 %	
 %	Start persistent operations using Directory   as  place to store
@@ -133,7 +136,7 @@ rdf_attach_db(DirSpec, Options) :-
 	    assert(rdf_directory(Directory)),
 	    assert_options(Options),
 	    stop_monitor,		% make sure not to register load
-	    load_db,
+	    no_agc(load_db),
 	    at_halt(rdf_detach_db),
 	    start_monitor
 	).
@@ -154,7 +157,7 @@ assert_options([H|T]) :-
 	(   option_type(H, Check)
 	->  Check,
 	    assert(rdf_option(H))
-	;   domain_error(rdf_option, H)
+	;   true			% ignore options we do not understand
 	),
 	assert_options(T).
 	
@@ -164,6 +167,18 @@ option_type(silent(X),	       must_be(oneof([true,false,brief]), X)).
 option_type(log_nested_transactions(X),	must_be(boolean, X)).
 
 
+%%	no_agc(:Goal)
+%
+%	Run Goal with atom garbage collection   disabled. Loading an RDF
+%	database creates large amounts  of  atoms   we  *know*  are  not
+%	garbage. 
+
+no_agc(Goal) :-
+	current_prolog_flag(agc_margin, Old),
+	set_prolog_flag(agc_margin, 0),
+	call_cleanup(Goal, set_prolog_flag(agc_margin, Old)).
+
+
 %%	rdf_detach_db is det.
 %
 %	Detach from the  current  database.   Succeeds  silently  if  no
@@ -333,7 +348,7 @@ load_source(DB, Silent, Nth, Total) :-
 	db_files(DB, SnapshotFile, JournalFile),
 	rdf_retractall(_,_,_,DB),
 	statistics(cputime, T0),
-	print_message(Level, rdf(restore(Silent, source(DB)))),
+	print_message(Level, rdf(restore(Silent, source(DB, Nth, Total)))),
 	db_file(SnapshotFile, AbsSnapShot),
 	(   exists_file(AbsSnapShot)
 	->  print_message(Level, rdf(restore(Silent, snapshot(SnapshotFile)))),
@@ -1071,14 +1086,15 @@ message(update_failed(S,P,O,Action)) -->
 
 silent_message(_Action) --> [].
 
-brief_message(source(_DB))     --> [].
-brief_message(snapshot(_File)) --> [].
-brief_message(journal(_File))  --> [].
-brief_message(done(_DB, _Time, _Count, Nth, Total)) -->
+brief_message(source(DB, Nth, Total)) -->
+	{ file_base_name(DB, Base) },
 	[ at_same_line, 
-	  '\rloaded ~|~t~D~5+ of ~|~t~D~5+ graphs'-[Nth, Total],
+	  '\r~w~`.t ~D of ~D graphs~72|'-[Base, Nth, Total],
 	  flush
 	].
+brief_message(snapshot(_File)) --> [].
+brief_message(journal(_File))  --> [].
+brief_message(done(_DB, _Time, _Count, _Nth, _Total)) --> [].
 
 prolog:message_context(rdf_locked(Args)) -->
 	{ memberchk(time(Time), Args),
