@@ -122,11 +122,21 @@ class_variable(size,	size,	size(200, 500),
 variable(file_pattern,	regex, get, "Pattern of showed files").
 
 initialise(FB, Root:directory) :->
-	send(FB, slot, file_pattern, '.*\.pl$'),
+	source_pattern(Regex),
+	send(FB, slot, file_pattern, Regex),
 	send_super(FB, initialise, Root),
 	send(FB?frame, label, 'SWI-Prolog Navigator'),
 	asserta(prolog_overview_window(FB)).
 
+source_pattern(Pat) :-
+	findall(E, (prolog_file_type(E, prolog),
+		    \+ prolog_file_type(E, qlf)), Exts),
+	(   Exts = [Ext]
+	->  format(atom(Pat), '.*\\.~w$', [Ext])
+	;   concat_atom(Exts, '|', P1),
+	    format(atom(Pat), '.*\\.(~w)$', [P1])
+	).
+
 unlink(FB) :->
 	retractall(prolog_overview_window(FB)),
 	send_super(FB, unlink).
@@ -186,8 +196,10 @@ goto(FB, File:'file|node', Line:int) :->
 		    send(FB, normalise, Node)
 		;   send(FB, goto, N2, Line)
 		)
-	    ;	get(Node?sons, tail, Last),
-		send(FB, goto, Last, Line)
+	    ;	get(Node?sons, tail, Last)
+	    ->	send(FB, goto, Last, Line)
+	    ;	send(FB?tree, selection, Node),
+		send(FB, normalise, Node)
 	    )
 	).
 
