@@ -1,4 +1,4 @@
-/*  $Id: save.c,v 1.20 2005/04/07 12:12:08 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -485,6 +485,7 @@ storeSlotsClass(Class class, FileObj file)
   succeed;
 }
 
+
 		/********************************
 		*            LOADING            *
 		*********************************/
@@ -524,6 +525,30 @@ loadWord(IOSTREAM *fd)
 }
 
 
+#ifdef WORDS_BIGENDIAN
+static const int double_byte_order[] = { 7,6,5,4,3,2,1,0 };
+#else
+static const int double_byte_order[] = { 0,1,2,3,4,5,6,7 };
+#endif
+
+#define BYTES_PER_DOUBLE (sizeof(double_byte_order)/sizeof(int))
+
+double
+loadDouble(IOSTREAM *fd)
+{ double f;
+  unsigned char *cl = (unsigned char *)&f;
+  unsigned int i;
+
+  for(i=0; i<BYTES_PER_DOUBLE; i++)
+  { int c = Sgetc(fd);
+    
+    cl[double_byte_order[i]] = c;
+  }
+  
+  return f;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 loadStringFile() loads a string saved using storeStringFile().  See there for
 format details.
@@ -538,7 +563,7 @@ loadStringFile(IOSTREAM *fd, String s)
     s->size = size;
 
     str_alloc(s);
-    if ( Sfread(s->s_textA, sizeof(char), size, fd) != size )
+    if ( Sfread(s->s_textA, sizeof(char), size, fd) != (size_t)size )
       fail;
   } else
   { int i;
@@ -666,12 +691,12 @@ status
 checkObjectMagic(IOSTREAM *fd)
 { status rval;
   long l;
-  int ls;
+  long ls;
 
   if ( SaveMagic == NULL )
     SaveMagic = SAVEMAGIC;
 
-  ls = strlen(SaveMagic);
+  ls = (long)strlen(SaveMagic);
 
   if ( (l=loadWord(fd)) == ls )
   { char tmp[LINESIZE];
