@@ -1,4 +1,4 @@
-/*  $Id: msfont.c,v 1.13 2005/01/03 11:28:50 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -229,7 +229,7 @@ parse_font(char *s, LOGFONT *lfont)
 status
 ws_create_font(FontObj f, DisplayObj d)
 { WsFont wsf = alloc(sizeof(ws_font));
-#ifdef __WIN32__
+#ifdef __WINDOWS__
   int widths[FONTTABLESIZE];
 #else
   short widths[FONTTABLESIZE];
