@@ -1,4 +1,4 @@
-/*  $Id: mswin.c,v 1.40 2006/01/18 12:32:17 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -235,6 +235,35 @@ ws_user()
 }
 
 
+#include <Shlobj.h>
+
+Name
+ws_appdata(const char *sub)
+{ TCHAR buf[MAX_PATH];
+
+  if ( SHGetSpecialFolderPath(0, buf, CSIDL_APPDATA, TRUE) )
+  { wchar_t *p;
+
+    for(p=buf; *p; p++)
+    { if ( *p == '\\' )
+	*p = '/';
+    }
+    if ( sub )
+    { const char *s;
+
+      *p++ = '/';
+      for(s=sub; *s; )
+	*p++ = *s++;
+      *p = EOS;
+    }
+
+    return TCHARToName(buf);
+  }
+
+  fail;
+}
+
+
 int
 ws_mousebuttons()
 { return GetSystemMetrics(SM_CMOUSEBUTTONS);
@@ -310,7 +339,7 @@ get_logical_drive_strings(int bufsize, char *buf)
 #define MAXPATHLEN _MAX_PATH
 #endif
 #define strapp(s, q) \
-	{ int l = _tcslen(q); \
+	{ size_t l = _tcslen(q); \
 	  if ( s+l+2 > filter+sizeof(filter)/sizeof(TCHAR) ) \
 	  { errorPce(filters, NAME_representation, NAME_nameTooLong); \
 	    fail; \
