@@ -129,7 +129,6 @@ struct arithFunction
 static ArithFunction	isCurrentArithFunction(functor_t, Module);
 static int		registerFunction(ArithFunction f, int index);
 static int		getCharExpression(term_t t, Number r ARG_LD);
-static int		ar_sign_i(Number n1);
 static int		ar_add(Number n1, Number n2, Number r);
 static int		ar_add_ui(Number n, long val);
 static int		ar_minus(Number n1, Number n2, Number r);
@@ -686,7 +685,7 @@ check_float(double f)
 
 
 int
-valueExpression(term_t t, Number r ARG_LD)
+eval_expression(term_t t, Number r, int recursion ARG_LD)
 { ArithFunction f;
   functor_t functor;
   Word p = valTermRef(t);
@@ -723,6 +722,9 @@ valueExpression(term_t t, Number r ARG_LD)
       return PL_error(NULL, 0, NULL, ERR_NOT_EVALUABLE, functor);
   }
 
+  if ( recursion > 100 && !PL_is_acyclic(t) )
+    return PL_error(NULL, 0, "cyclic term", ERR_TYPE, ATOM_expression, t);
+
 #if O_PROLOG_FUNCTIONS
   if ( f->proc )
   { int rval, n, arity = arityFunctor(functor);
@@ -733,7 +735,7 @@ valueExpression(term_t t, Number r ARG_LD)
     { number n1;
 
       _PL_get_arg(n+1, t, h0+n);
-      if ( valueExpression(h0+n, &n1 PASS_LD) )
+      if ( eval_expression(h0+n, &n1, recursion+1 PASS_LD) )
       { _PL_put_number(h0+n, &n1);
 	clearNumber(&n1);
       } else
@@ -771,7 +773,7 @@ valueExpression(term_t t, Number r ARG_LD)
 	number n1;
 
 	_PL_get_arg(1, t, a);
-	if ( valueExpression(a, &n1 PASS_LD) )
+	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
 	{ rval = (*f->function)(&n1, r);
 	  clearNumber(&n1);
 	} else
@@ -785,9 +787,9 @@ valueExpression(term_t t, Number r ARG_LD)
 	number n1, n2;
 
 	_PL_get_arg(1, t, a);
-	if ( valueExpression(a, &n1 PASS_LD) )
+	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
 	{ _PL_get_arg(2, t, a);
-	  if ( valueExpression(a, &n2 PASS_LD) )
+	  if ( eval_expression(a, &n2, recursion+1 PASS_LD) )
 	  { rval = (*f->function)(&n1, &n2, r);
 	    clearNumber(&n2);
 	  } else
@@ -805,11 +807,11 @@ valueExpression(term_t t, Number r ARG_LD)
 	number n1, n2, n3;
 
 	_PL_get_arg(1, t, a);
-	if ( valueExpression(a, &n1 PASS_LD) )
+	if ( eval_expression(a, &n1, recursion+1 PASS_LD) )
 	{ _PL_get_arg(2, t, a);
-	  if ( valueExpression(a, &n2 PASS_LD) )
+	  if ( eval_expression(a, &n2, recursion+1 PASS_LD) )
 	  { _PL_get_arg(3, t, a);
-	    if ( valueExpression(a, &n3 PASS_LD) )
+	    if ( eval_expression(a, &n3, recursion+1 PASS_LD) )
 	    { rval = (*f->function)(&n1, &n2, &n3, r);
 	      clearNumber(&n3);
 	    } else
@@ -852,6 +854,12 @@ valueExpression(term_t t, Number r ARG_LD)
 }
 
 
+int
+valueExpression(term_t t, Number r ARG_LD)
+{ return eval_expression(t, r, 0 PASS_LD);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 int arithChar(Word p)
     Handle arithmetic argument "x", normally appearing as [X], where X
@@ -1153,6 +1161,11 @@ ar_even(Number i)
 #endif
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+mod(X, Y) = X - (floor(X/Y) * Y)
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+
 static int
 ar_mod(Number n1, Number n2, Number r)
 { if ( !toIntegerNumber(n1, 0) )
@@ -1167,7 +1180,11 @@ ar_mod(Number n1, Number n2, Number r)
       if ( n2->value.i == 0 )
 	return PL_error("mod", 2, NULL, ERR_DIV_BY_ZERO);
 
-      r->value.i = n1->value.i % n2->value.i;
+      { int64_t mod = n1->value.i % n2->value.i;
+	if ( mod != 0 && ( mod < 0 ) != ( n2->value.i < 0 ) )
+	  mod += n2->value.i;
+	r->value.i = mod;
+      }
       r->type = V_INTEGER;
       break;
 #ifdef O_GMP
@@ -1177,9 +1194,7 @@ ar_mod(Number n1, Number n2, Number r)
 
       r->type = V_MPZ;
       mpz_init(r->value.mpz);
-      mpz_mod(r->value.mpz, n1->value.mpz, n2->value.mpz);
-      if ( mpz_sgn(n1->value.mpz) < 0 )
-	mpz_neg(r->value.mpz, r->value.mpz);
+      mpz_fdiv_r(r->value.mpz, n1->value.mpz, n2->value.mpz);
       break;
 #endif
     default:
@@ -1599,7 +1614,7 @@ ar_div(Number n1, Number n2, Number r)
 }
 
 
-static int
+int
 ar_sign_i(Number n1)
 { switch(n1->type)
   { case V_INTEGER:
@@ -1628,9 +1643,7 @@ ar_sign(Number n1, Number r)
 
 static int
 ar_rem(Number n1, Number n2, Number r)
-{ real f;
-
-  if ( !toIntegerNumber(n1, 0) )
+{ if ( !toIntegerNumber(n1, 0) )
     return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
   if ( !toIntegerNumber(n2, 0) )
     return PL_error("rem", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
@@ -1638,21 +1651,21 @@ ar_rem(Number n1, Number n2, Number r)
   same_type_numbers(n1, n2);
   switch(n1->type)
   { case V_INTEGER:
-      f = (real)n1->value.i / (real)n2->value.i;
-      r->value.f = f - (real)((int64_t) f);
-      r->type = V_REAL;
+      if ( n2->value.i == 0 )
+	return PL_error("rem", 2, NULL, ERR_DIV_BY_ZERO);
+
+      r->value.i = n1->value.i % n2->value.i;
+      r->type = V_INTEGER;
+
       break;
 #ifdef O_GMP
     case V_MPZ:
-    { mpq_t q;
-      
-      mpq_init(q);
-      mpz_mod(mpq_numref(q), n1->value.mpz, n2->value.mpz);
-      mpz_set(mpq_denref(q), n2->value.mpz);
+    { if ( mpz_sgn(n2->value.mpz) == 0 )
+	return PL_error("rem", 2, NULL, ERR_DIV_BY_ZERO);
 
-      r->value.f = mpq_get_d(q);
-      r->type = V_REAL;
-      mpq_clear(q);
+      r->type = V_MPZ;
+      mpz_init(r->value.mpz);
+      mpz_tdiv_r(r->value.mpz, n1->value.mpz, n2->value.mpz);
       break;
     }
 #endif
@@ -2209,6 +2222,14 @@ ar_u_plus(Number n1, Number r)
 }
 
 
+static int
+ar_eval(Number n1, Number r)
+{ cpNumber(r, n1);
+
+  succeed;
+}
+
+
 static int
 ar_abs(Number n1, Number r)
 { switch(n1->type)
@@ -2849,7 +2870,9 @@ static const ar_funcdef ar_funcdefs[] = {
   ADD(FUNCTOR_msb1,		ar_msb),
   ADD(FUNCTOR_lsb1,		ar_lsb),
   ADD(FUNCTOR_popcount1,	ar_popcount),
-  ADD(FUNCTOR_powm3,		ar_powm)
+  ADD(FUNCTOR_powm3,		ar_powm),
+
+  ADD(FUNCTOR_eval1,		ar_eval)
 };
 
 #undef ADD
@@ -3024,13 +3047,13 @@ ar_func_n(int findex, int argc ARG_LD)
 		 *******************************/
 
 BeginPredDefs(arith)
-  PRED_DEF("is",   2, is,  PL_FA_TRANSPARENT)
-  PRED_DEF("<",	   2, lt,  PL_FA_TRANSPARENT)
-  PRED_DEF(">",	   2, gt,  PL_FA_TRANSPARENT)
-  PRED_DEF("=<",   2, leq, PL_FA_TRANSPARENT)
-  PRED_DEF(">=",   2, geq, PL_FA_TRANSPARENT)
-  PRED_DEF("=\\=", 2, neq, PL_FA_TRANSPARENT)
-  PRED_DEF("=:=",  2, eq,  PL_FA_TRANSPARENT)
+  PRED_DEF("is",   2, is,  PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF("<",	   2, lt,  PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF(">",	   2, gt,  PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF("=<",   2, leq, PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF(">=",   2, geq, PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF("=\\=", 2, neq, PL_FA_TRANSPARENT|PL_FA_ISO)
+  PRED_DEF("=:=",  2, eq,  PL_FA_TRANSPARENT|PL_FA_ISO)
   PRED_DEF("$prolog_arithmetic_function", 2, prolog_arithmetic_function,
 	   PL_FA_NONDETERMINISTIC|PL_FA_TRANSPARENT)
   PRED_DEF("$arithmetic_function", 2, arithmetic_function, PL_FA_TRANSPARENT)
