@@ -119,6 +119,8 @@ char tmp[256];				/* for calling print_val(), etc. */
 #define recordMark(p)
 #define needsRelocation(p) { needs_relocation++; }
 #define check_relocation(p)
+#define markLocal(p) (local_marked++)
+#define processLocal(p) (local_marked--)
 #endif
 
 #define ldomark(p)	{ *(p) |= MARK_MASK; }
@@ -209,6 +211,7 @@ static int		check_marked(const char *s);
 #define mark_base	   (LD->gc._mark_base)
 #define mark_top	   (LD->gc._mark_top)
 #define check_table	   (LD->gc._check_table)
+#define local_table	   (LD->gc._local_table)
 #endif
 
 #undef LD
@@ -313,6 +316,28 @@ do_check_relocation(Word addr, char *file, int line ARG_LD)
   s->value = FALSE;
 }
 
+
+static void
+markLocal(Word addr)
+{ GET_LD
+  
+  local_marked++;
+  addHTable(local_table, addr, (Void)TRUE);
+}
+
+static void
+processLocal(Word addr)
+{ GET_LD
+  Symbol s;
+
+  local_marked--;
+  if ( (s = lookupHTable(local_table, addr)) )
+  { s->value = (Void)FALSE;
+  } else
+  { assert(0);
+  }
+}
+
 #endif /* O_SECURE */
 
 		/********************************
@@ -457,11 +482,13 @@ mark_variable(Word start ARG_LD)
   if ( is_marked(start) )
     sysError("Attempt to mark twice");
 
-  local_marked++;
+  if ( onStackArea(local, start) )
+  { markLocal(start);
+    total_marked--;			/* do not count local stack cell */
+  }
   current = start;
   mark_first(current);
   val = get_value(current);  
-  total_marked--;			/* do not count local stack cell */
   FORWARD;
 
 forward:				/* Go into the tree */
@@ -672,7 +699,7 @@ gvars_to_term_refs(Word **saved_bar_at)
     freeTableEnum(e);
     assert(LD->gvar.grefs == found);
 
-    DEBUG(0, Sdprintf("Found %d global vars on global stack. "
+    DEBUG(1, Sdprintf("Found %d global vars on global stack. "
 		      "stored in frame %p\n", found, fli_context));
   } else
     fid = 0;
@@ -741,9 +768,7 @@ mark_attvars()
 
   for( gp = gBase; gp < gTop; gp += (offset_cell(gp)+1) )
   { if ( isAttVar(*gp) && !is_marked(gp) )
-    { local_marked--;			/* see mark_variable() */
-      total_marked++;
-      DEBUG(3, Sdprintf("mark_attvars(): marking %p\n", gp));
+    { DEBUG(3, Sdprintf("mark_attvars(): marking %p\n", gp));
       mark_variable(gp PASS_LD);
     }
   }
@@ -1009,10 +1034,7 @@ early_reset_vars(mark *m, Word top, GCTrailEntry te ARG_LD)
 	assert(onGlobal(gp));
 	assert(!is_first(gp));
 	if ( !is_marked(gp) )
-	{ total_marked++;			/* fix counters */
-	  local_marked--;
-
-	  DEBUG(2,
+	{ DEBUG(3,
 		char b1[64]; char b2[64]; char b3[64];
 		Sdprintf("Marking assignment at %s (%s --> %s)\n",
 			 print_adr(tard, b1),
@@ -1373,17 +1395,17 @@ tag_trail()
 
       SECURE(assert(onStack(global, p2)));
       te->address = (Word)consPtr(p2, STG_GLOBAL|TAG_TRAILVAL);
-      te--;
-    }
-
-    if ( onLocal(te->address) )
-    { stg = STG_LOCAL;
+      //SECURE(assert(te == tBase || !isTrailVal(te[-1].address)));
     } else
-    { SECURE(assert(onStackArea(global, te->address)));
-      stg = STG_GLOBAL;
+    { if ( onLocal(te->address) )
+      { stg = STG_LOCAL;
+      } else
+      { SECURE(assert(onStackArea(global, te->address)));
+	stg = STG_GLOBAL;
+      }
+	
+      te->address = (Word)consPtr(te->address, stg);
     }
-
-    te->address = (Word)consPtr(te->address, stg);
   }
 }
 
@@ -1462,7 +1484,7 @@ sweep_foreign()
     { if ( is_marked(sp) )
       {	unmark(sp);
 	if ( isGlobalRef(get_value(sp)) )
-	{ local_marked--;
+	{ processLocal(sp);
 	  check_relocation(sp);
 	  into_relocation_chain(sp, STG_LOCAL PASS_LD);
 	}
@@ -1569,7 +1591,7 @@ sweep_environments(LocalFrame fr, Code PC)
     { if ( is_marked(sp) )
       { unmark(sp);
 	if ( isGlobalRef(get_value(sp)) )
-	{ local_marked--;
+	{ processLocal(sp);
 	  check_relocation(sp);
 	  into_relocation_chain(sp, STG_LOCAL PASS_LD);
 	}
@@ -1621,7 +1643,27 @@ sweep_stacks(LocalFrame fr, Choice ch)
   }
 
   if ( local_marked != 0 )
+  {
+#ifdef O_SECURE
+    TableEnum e = newTableEnum(local_table);
+    Symbol s;
+
+    Sdprintf("FATAL: unprocessed local variables:\n");
+
+    while((s=advanceTableEnum(e)))
+    { if ( s->value )
+      {	Word p = s->name;
+	char buf1[64];
+	char buf2[64];
+
+	Sdprintf("\t%s (*= %s)\n", print_adr(p, buf1), print_val(*p, buf2));
+      }
+    }
+
+    freeTableEnum(e);
+#endif
     sysError("local_marked = %ld", local_marked);
+  }
 }
 
 
@@ -2208,11 +2250,14 @@ garbageCollect(LocalFrame fr, Choice ch)
   sigset_t mask;
   fid_t fid;
   Word *saved_bar_at;
+#ifdef O_PROFILE
+  struct call_node *prof_node = NULL;
+#endif
 #ifdef O_SECURE
   word key;
 #endif
 
-  DEBUG(0, verbose = TRUE);
+  DEBUG(1, verbose = TRUE);
 
   if ( gc_status.blocked || !trueFeature(GC_FEATURE) )
     return;
@@ -2237,7 +2282,7 @@ garbageCollect(LocalFrame fr, Choice ch)
 
 #ifdef O_PROFILE
   if ( LD->profile.active )
-    profCall(PROCEDURE_garbage_collect0->definition PASS_LD);
+    prof_node = profCall(PROCEDURE_garbage_collect0->definition PASS_LD);
 #endif
 
 #if O_SECURE
@@ -2247,9 +2292,12 @@ garbageCollect(LocalFrame fr, Choice ch)
   key = checkStacks(fr, ch);
 
   if ( check_table == NULL )
-    check_table = newHTable(256);
-  else
-    clearHTable(check_table);
+  { check_table = newHTable(256);
+    local_table = newHTable(256);
+  } else
+  { clearHTable(check_table);
+    clearHTable(local_table);
+  }
 
   mark_base = mark_top = malloc(usedStack(global));
 #endif
@@ -2319,8 +2367,8 @@ garbageCollect(LocalFrame fr, Choice ch)
 		     PL_INTPTR, roomStack(trail));
 
 #ifdef O_PROFILE
-  if ( fr && LD->profile.active )
-    profExit(fr->prof_node PASS_LD);
+  if ( prof_node && LD->profile.active )
+    profExit(prof_node PASS_LD);
 #endif
 
   unblockGC(PASS_LD1);
@@ -2818,7 +2866,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC,
     double time = CpuTime(CPU_USER);
     int verbose = trueFeature(TRACE_GC_FEATURE);
     
-    DEBUG(0, verbose = TRUE);
+    DEBUG(1, verbose = TRUE);
 
     if ( verbose )
     { printMessage(ATOM_informational,
@@ -2885,7 +2933,7 @@ growStacks(LocalFrame fr, Choice ch, Code PC,
 	}
 
 
-    DEBUG(0, { Sputchar('\n');
+    DEBUG(1, { Sputchar('\n');
 	       PrintStackParms(global, "global", gb, gsize);
 	       PrintStackParms(local, "local", lb, lsize);
 	       PrintStackParms(trail, "trail", tb, tsize);
